/* x11 video driver implementation */

#include "compat.h"
#include "console.h"
#include "console_struct.h"
#include "vt_buffer.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>

static Display *display;
static int screen_num;
static Window win;
static XFontStruct *font_info;
static GC gc, cursorgc;
static int charwidth, charheight;
static int cursor_x = -1, cursor_y = -1;
#define XPOS(x) ((x)*charwidth)
#define YPOS(y) ((y)*charheight)
#define TYPOS(y) ((y)*charheight + font_info->max_bounds.ascent)
#define XSIZE(x) ((x)*charwidth)
#define YSIZE(y) ((y)*charheight)
#define LOCATION(y, x) (&(c->vc_screenbuf[c->vc_cols*(y) + (x)]))
#define MIN(x, y) ((x) < (y) ? (x) : (y))

static void die(char *msg) {
  fprintf(stderr, "error: %s\n", msg);
  exit(20);
}

static void x11_init(struct vc_data *c, int init) {
  int x, y;
  unsigned int width, height, border_width = 2;
  XGCValues values;
  XColor cursorcol, dummycol;

  c->vc_cols = 80;
  c->vc_rows = 25;

  display = XOpenDisplay(NULL);
  if (!display) die("cannot open display");
  screen_num = DefaultScreen(display);

  font_info = XLoadQueryFont(display, "fixed");
  if (!font_info) die("cannot open font");

  x = y = 0;
  charwidth = font_info->max_bounds.width;
  charheight = font_info->max_bounds.ascent + font_info->max_bounds.descent;
  width = XSIZE(c->vc_cols);
  height = YSIZE(c->vc_rows);
  
  win = XCreateSimpleWindow(display, RootWindow(display, screen_num),
			    x, y, width, height, border_width,
			    BlackPixel(display, screen_num),
			    WhitePixel(display, screen_num));

  gc = XCreateGC(display, win, 0, &values);
  XSetFont(display, gc, font_info->fid);
  XSetForeground(display, gc, BlackPixel(display, screen_num));
  XSetBackground(display, gc, WhitePixel(display, screen_num));

  XAllocNamedColor(display, DefaultColormap(display, screen_num),
		   "red", &cursorcol, &dummycol);

  values.function = GXinvert;
  cursorgc = XCreateGC(display, win, GCFunction, &values);
 
  XSelectInput(display, win, ExposureMask | KeyPressMask);
  
  XMapWindow(display, win);
}

static void x11_deinit(struct vc_data *c) {
  XUnloadFont(display, font_info->fid);
  XFreeGC(display, gc);
  XCloseDisplay(display);
}

static void x11_invert_region(struct vc_data *c, u16 *p, int count) {
  printf("invert region\n");
}

static void draw(struct vc_data *c, int ch, int y, int x, GC context) {
  char c8 = ch & 0xff;
  XDrawImageString(display, win, context, XPOS(x), TYPOS(y), &c8, 1);
}

static void draws(struct vc_data *c, const unsigned short *s, int count,
		  int y, int x, GC context) {
  int i;
  /* inefficient, should convert and draw */
  for (i=0; i<count; i++) draw(c, s[i], y, x+i, context);
}

static void draw_cursor(struct vc_data *c) {
#warning should check if the cursor is visible
  XFillRectangle(display, win, cursorgc, XPOS(cursor_x), YPOS(cursor_y),
		 XSIZE(1), YSIZE(1));
}

static void redraw(struct vc_data *c, int y, int x, int len) {
  draws(c, LOCATION(y, x), len, y, x, gc);
}

static void redraw_all(struct vc_data *c) {
  int i;
  for (i=0; i<c->vc_rows; i++) redraw(c, i, 0, c->vc_cols);
  draw_cursor(c);
}

static void update_cursor(struct vc_data *c) {
  cursor_x = c->vc_x;
  cursor_y = c->vc_y;
}

static void x11_putc(struct vc_data *c, int ch, int y, int x) {
  /* should look at ch&0xff00 to pick the gc */
  draw_cursor(c);
  draw(c, ch, y, x, gc);
  update_cursor(c);
  draw_cursor(c);
}

static void x11_putcs(struct vc_data *c, const unsigned short *s,
		      int count, int y, int x) {
  draw_cursor(c);
  draws(c, s, count, y, x, gc);
  update_cursor(c);
  draw_cursor(c);
}

static void x11_clear(struct vc_data *c, int y, int x, 
		      int height, int width) {
  draw_cursor(c);
  XClearArea(display, win, XPOS(x), YPOS(y),
	     XSIZE(width), YSIZE(height), False);
  update_cursor(c);
  draw_cursor(c);
}
                        
static void x11_bmove(struct vc_data *c, int sy, int sx, 
		      int dy, int dx, int height, int width) {
#warning check whether this should clear
  draw_cursor(c);
  XCopyArea(display, win, win, gc,
	    XPOS(sx), YPOS(sy),
	    XSIZE(width), YSIZE(height),
	    XPOS(dx), YPOS(dy));
  update_cursor(c);
  draw_cursor(c);
}

static int x11_set_palette(struct vc_data *c, unsigned char *table) {
  printf("set palette\n");

  /* The return code does not appear to be used. */
  return 0;
}

static void x11_cursor(struct vc_data *c, int mode) {
  printf("cursor %d\n", mode);
  draw_cursor(c);
  update_cursor(c);
  draw_cursor(c);
}

static int x11_scroll(struct vc_data *c, int t, int b, int dir, int lines) {
  int i;

  draw_cursor(c);

#warning this is completely wrong and needs to check dir
  for (i=t; i<b; i++) {
    x11_clear(c, i, 0, 1, c->vc_cols);
    x11_bmove(c, i+1, 0, i, 0, 1, c->vc_cols);
  }
  x11_clear(c, i, 0, 1, c->vc_cols);

  update_cursor(c);
  draw_cursor(c);

  return 1;
}

#define KEYBUF_SIZE 256
static char keybuf[KEYBUF_SIZE];

static int get_key(XEvent *ev) {
  KeySym keysym;
  int count;

  count = XLookupString(&ev->xkey, keybuf, KEYBUF_SIZE, &keysym, NULL);
#warning convert metas etc.
  keybuf[count] = '\0';
  return count;
}

static int send_count = 0, send_offset = 0;
static char *send_buf;

static void send_str(char *data, int count) {
  if (send_count) {
    send_buf = realloc(send_buf, send_count + count);
    memcpy(send_buf + send_count, data, count);
    send_count += count;
  } else {
    send_buf = malloc(count);
    memcpy(send_buf, data, count);
    send_count = count;
    send_offset = 0;
  }
}

static void x11_mainloop(struct vc_data *c, int ptyfd) {
  int xfd = XConnectionNumber(display);
  fd_set infds, outfds;

  while (1) {
    FD_ZERO(&infds);
    FD_ZERO(&outfds);
    FD_SET(ptyfd, &infds);
    FD_SET(xfd, &infds);
    if (send_count) FD_SET(ptyfd, &outfds);
    
    if (FD_ISSET(xfd, &infds)) {
      XEvent ev;

      while (XPending(display) > 0) {
	int count;

	XNextEvent(display, &ev);
	switch (ev.type) {
	case Expose:
	  /* If we get a sequence of Expose events, there's no point
	     in redrawing the window for each of them. */
	  if (ev.xexpose.count) break;
	  redraw_all(c);
	  break;
	case ConfigureNotify:
	  printf("got configure event\n");
	  break;
	case KeyPress:
	  count = get_key(&ev);
	  send_str(keybuf, count);
	  break;
	default:
	  //printf("got other event\n");
	  break;
	}
      }
    }

    if (FD_ISSET(ptyfd, &infds)) {
#define READ_CHUNKSIZE 512
      char readbuf[READ_CHUNKSIZE];
      int count;
      count = read(ptyfd, readbuf, READ_CHUNKSIZE);
      if (count < 0) {
	if (errno == EWOULDBLOCK) goto readout;
	perror("pty");
	die("pty read");
      }
      do_con_write(readbuf, count);
    }
  readout:

    if (FD_ISSET(ptyfd, &outfds)) {
#define WRITE_CHUNKSIZE 512
      int count;
      count = write(ptyfd, send_buf + send_offset,
		    MIN(send_count, WRITE_CHUNKSIZE));
      if (count < 0) {
	perror("pty");
	die("pty write");
      }
      send_offset += count;
      send_count -= count;
      if (!send_count) free(send_buf);
    }
  }
}

const struct consw x11_con = {
  con_init: x11_init,
  con_deinit: x11_deinit,
  con_mainloop: x11_mainloop,
  con_clear: x11_clear,
  con_putc: x11_putc,
  con_putcs: x11_putcs,
  con_cursor: x11_cursor,
  con_scroll: x11_scroll,
  con_bmove: x11_bmove,
  con_set_palette: x11_set_palette,
  con_build_attr: NULL,
  con_invert_region: x11_invert_region
};
